#pragma once

#define FF_COLOR_MODE_RESET "0;"
#define FF_COLOR_MODE_BOLD "1;"
#define FF_COLOR_MODE_DIM "2;"
#define FF_COLOR_MODE_ITALIC "3;"
#define FF_COLOR_MODE_UNDERLINE "4;"
#define FF_COLOR_MODE_BLINK "5;"
#define FF_COLOR_MODE_INVERSE "7;"
#define FF_COLOR_MODE_HIDDEN "8;"
#define FF_COLOR_MODE_STRIKETHROUGH "9;"

#define FF_COLOR_FG_BLACK "30"
#define FF_COLOR_FG_RED "31"
#define FF_COLOR_FG_GREEN "32"
#define FF_COLOR_FG_YELLOW "33"
#define FF_COLOR_FG_BLUE "34"
#define FF_COLOR_FG_MAGENTA "35"
#define FF_COLOR_FG_CYAN "36"
#define FF_COLOR_FG_WHITE "37"
#define FF_COLOR_FG_DEFAULT "39"

#define FF_COLOR_FG_LIGHT_BLACK "90"
#define FF_COLOR_FG_LIGHT_RED "91"
#define FF_COLOR_FG_LIGHT_GREEN "92"
#define FF_COLOR_FG_LIGHT_YELLOW "93"
#define FF_COLOR_FG_LIGHT_BLUE "94"
#define FF_COLOR_FG_LIGHT_MAGENTA "95"
#define FF_COLOR_FG_LIGHT_CYAN "96"
#define FF_COLOR_FG_LIGHT_WHITE "97"

#define FF_COLOR_BG_BLACK "40"
#define FF_COLOR_BG_RED "41"
#define FF_COLOR_BG_GREEN "42"
#define FF_COLOR_BG_YELLOW "43"
#define FF_COLOR_BG_BLUE "44"
#define FF_COLOR_BG_MAGENTA "45"
#define FF_COLOR_BG_CYAN "46"
#define FF_COLOR_BG_WHITE "47"
#define FF_COLOR_BG_DEFAULT "49"

#define FF_COLOR_BG_LIGHT_BLACK "100"
#define FF_COLOR_BG_LIGHT_RED "101"
#define FF_COLOR_BG_LIGHT_GREEN "102"
#define FF_COLOR_BG_LIGHT_YELLOW "103"
#define FF_COLOR_BG_LIGHT_BLUE "104"
#define FF_COLOR_BG_LIGHT_MAGENTA "105"
#define FF_COLOR_BG_LIGHT_CYAN "106"
#define FF_COLOR_BG_LIGHT_WHITE "107"

#define FF_COLOR_FG_256 "38;5;"
#define FF_COLOR_BG_256 "48;5;"

#define FF_COLOR_FG_RGB "38;2;"
#define FF_COLOR_BG_RGB "48;2;"
