#pragma once

// For "fastfetch.h"

#include "modules/battery/option.h"
#include "modules/bios/option.h"
#include "modules/bluetooth/option.h"
#include "modules/bluetoothradio/option.h"
#include "modules/board/option.h"
#include "modules/bootmgr/option.h"
#include "modules/break/option.h"
#include "modules/brightness/option.h"
#include "modules/btrfs/option.h"
#include "modules/camera/option.h"
#include "modules/chassis/option.h"
#include "modules/cpu/option.h"
#include "modules/cpucache/option.h"
#include "modules/cpuusage/option.h"
#include "modules/colors/option.h"
#include "modules/cursor/option.h"
#include "modules/custom/option.h"
#include "modules/command/option.h"
#include "modules/datetime/option.h"
#include "modules/de/option.h"
#include "modules/disk/option.h"
#include "modules/diskio/option.h"
#include "modules/display/option.h"
#include "modules/dns/option.h"
#include "modules/editor/option.h"
#include "modules/font/option.h"
#include "modules/host/option.h"
#include "modules/gamepad/option.h"
#include "modules/gpu/option.h"
#include "modules/icons/option.h"
#include "modules/initsystem/option.h"
#include "modules/kernel/option.h"
#include "modules/keyboard/option.h"
#include "modules/loadavg/option.h"
#include "modules/locale/option.h"
#include "modules/lm/option.h"
#include "modules/localip/option.h"
#include "modules/media/option.h"
#include "modules/memory/option.h"
#include "modules/monitor/option.h"
#include "modules/mouse/option.h"
#include "modules/netio/option.h"
#include "modules/opengl/option.h"
#include "modules/opencl/option.h"
#include "modules/os/option.h"
#include "modules/packages/option.h"
#include "modules/physicaldisk/option.h"
#include "modules/physicalmemory/option.h"
#include "modules/player/option.h"
#include "modules/poweradapter/option.h"
#include "modules/processes/option.h"
#include "modules/publicip/option.h"
#include "modules/separator/option.h"
#include "modules/shell/option.h"
#include "modules/sound/option.h"
#include "modules/swap/option.h"
#include "modules/terminal/option.h"
#include "modules/terminalfont/option.h"
#include "modules/terminalsize/option.h"
#include "modules/terminaltheme/option.h"
#include "modules/theme/option.h"
#include "modules/title/option.h"
#include "modules/tpm/option.h"
#include "modules/uptime/option.h"
#include "modules/users/option.h"
#include "modules/version/option.h"
#include "modules/vulkan/option.h"
#include "modules/wallpaper/option.h"
#include "modules/weather/option.h"
#include "modules/wifi/option.h"
#include "modules/wm/option.h"
#include "modules/wmtheme/option.h"
#include "modules/zpool/option.h"
