#ifdef RC_INVOKED

#include <winuser.h>
#include <winver.h>
#include <ntdef.h>
#include "fastfetch_config.h"

#define FF_TO_STR1(str) #str
#define FF_TO_STR(str) FF_TO_STR1(str)

CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "manifest.xml"
id ICON "logo.ico"

VS_VERSION_INFO VERSIONINFO
    FILEVERSION             FASTFETCH_PROJECT_VERSION_MAJOR,FASTFETCH_PROJECT_VERSION_MINOR,FASTFETCH_PROJECT_VERSION_PATCH,FASTFETCH_PROJECT_VERSION_TWEAK_NUM
    PRODUCTVERSION          FASTFETCH_PROJECT_VERSION_MAJOR,FASTFETCH_PROJECT_VERSION_MINOR,FASTFETCH_PROJECT_VERSION_PATCH,FASTFETCH_PROJECT_VERSION_TWEAK_NUM
    FILEOS                  VOS_NT
    FILETYPE                VFT_APP
{
    BLOCK "StringFileInfo" {
        BLOCK "040904b0" {
            VALUE "Comments",         FASTFETCH_PROJECT_DESCRIPTION
            VALUE "FileDescription",  FF_TO_STR(FASTFETCH_TARGET_BINARY_NAME)
            VALUE "FileVersion",      FASTFETCH_PROJECT_VERSION FASTFETCH_PROJECT_VERSION_TWEAK
            VALUE "InternalName",     FF_TO_STR(FASTFETCH_TARGET_BINARY_NAME)
            VALUE "LegalCopyright",   FASTFETCH_PROJECT_LICENSE
            VALUE "OriginalFilename", FF_TO_STR(FASTFETCH_TARGET_BINARY_NAME) ".exe"
            VALUE "ProductName",      FASTFETCH_PROJECT_NAME " - " FASTFETCH_PROJECT_DESCRIPTION
            VALUE "ProductVersion",   FASTFETCH_PROJECT_VERSION FASTFETCH_PROJECT_VERSION_TWEAK
            VALUE "CompanyName",      FASTFETCH_PROJECT_HOMEPAGE_URL
        }
    }
    BLOCK "VarFileInfo" {
        VALUE "Translation", 0x0409, 1200
    }
}
#endif
